#include <stdio.h>
#include <assert.h>
#include <string.h>
#include "globus_common.h"
#include "globus_gara_common.h"
#include "globus_gara_client.h"
#include "globus_gara_end2end.h"

#define AUTHORIZATION_FAILED 2
#define RESERVATION_FAILED   3 

struct res_params_s
{
  char * host1;
  char * host2;
  int  s_time_offset;  /* seconds from "now" */
  int  duration;       /* seconds */
  int  bandwidth;      /* kbp's */
  char * protocol;     /* udp or tcp */
};
typedef struct res_params_s res_params_t;

static void process_arguments(res_params_t  *);
static char *make_reservation_rsl(res_params_t *);
static void check_status(char *);
static void callback(char *, int , void  *);

int main(int argc, char *argv[]) {

  int                        i;
  int                        error;
  int                        path_length;
  char                       **path;
  res_params_t               * params;
  char *x509_cert_dir=NULL, *x509_cert_file=NULL, 
       *x509_user_cert=NULL,*x509_user_key=argv[4], 
       *x509_user_proxy=NULL;
  char *operation=argv[6];
  char *rsl, *reservation_handle=argv[7];
  int size;

  x509_cert_dir = (char *)malloc(strlen("X509_CERT_DIR")+strlen(argv[1])+2);
  sprintf(x509_cert_dir, "%s=%s", "X509_CERT_DIR", argv[1]);
  fprintf(stderr, "%s\n", x509_cert_dir);
 
  x509_cert_file = (char *)malloc(strlen("X509_CERT_FILE")+strlen(argv[2])+2);
  sprintf(x509_cert_file, "%s=%s", "X509_CERT_FILE", argv[2]);
  fprintf(stderr, "%s\n", x509_cert_file);
 
  x509_user_cert = (char *)malloc(strlen("X509_USER_CERT")+strlen(argv[3])+2);
  sprintf(x509_user_cert, "%s %s", "X509_USER_CERT", argv[3]);
  fprintf(stderr, "%s\n", x509_user_cert);
 
  x509_user_key = (char *)malloc(strlen("X509_USER_KEY")+strlen(argv[4])+2);
  sprintf(x509_user_key, "%s %s", "X509_USER_KEY", argv[4]);
  fprintf(stderr, "%s\n", x509_user_key);
 
  x509_user_proxy = (char *)malloc(strlen("X509_USER_PROXY")+strlen(argv[5])+2);  
  sprintf(x509_user_proxy, "%s=%s", "X509_USER_PROXY", argv[5]);
  fprintf(stderr, "%s\n", x509_user_proxy);
                                     
  if(putenv(x509_cert_dir) != 0)
        fprintf(stderr, "unable to set %s\n", x509_cert_dir);
  if(putenv(x509_cert_file) != 0)
        fprintf(stderr, "unable to set %s\n", x509_cert_file);
  if(putenv(x509_user_cert) != 0 )
        fprintf(stderr, "unable to set %s\n", x509_user_cert);
  if(putenv(x509_user_key) != 0)
        fprintf(stderr, "unable to set %s\n", x509_user_key);
  if(putenv(x509_user_proxy) != 0)
        fprintf(stderr, "unable to set %s\n", x509_user_proxy); 

  if(!strncmp(operation, "create", 6)) {                
    if ((params = (res_params_t *)malloc(sizeof(res_params_t))) == NULL) {
      fprintf(stderr, "globus_malloc() failed\n");
      return 0;
    }
    params->host1 = (char *)malloc(strlen(argv[7])+1);
    strcpy(params->host1, argv[7]);
    params->host2 = (char *)malloc(strlen(argv[8])+1);
    strcpy(params->host2, argv[8]);
    params->s_time_offset = atoi(argv[9]); 
    params->duration = atoi(argv[10]);
    params->bandwidth = atoi(argv[11]);
    params->protocol = (char *)malloc(strlen(argv[12])+1);
    if(!strcmp(argv[12], "TCP"))
      strcpy(params->protocol, "tcp"); 
    else if(!strcmp(argv[12], "UDP"))
      strcpy(params->protocol, "udp" );
  }

  error = globus_module_activate(GLOBUS_GARA_END2END_MODULE);

  if (error)
  {
    fprintf(stderr, "Error on module_init(): %d\n", error);
    return error;
  }
 
  fprintf(stderr, "e2egara: operation = %s\n", operation);

  if(!strncmp(operation, "create", 6)) {
    rsl = make_reservation_rsl(params);

    fprintf(stderr, "About to call _reservation_create() with %s\n", rsl);
    
    error = globus_gara_end2end_reservation_create(rsl, &reservation_handle);
    fprintf(stderr, "error on _create = %d\n", error);
    switch(error) {
    case GLOBUS_GARA_ERROR_NONE: break;
    case GLOBUS_GARA_ERROR_AUTHORIZATION_FAILED:
      error = AUTHORIZATION_FAILED;
      break;
    default: 
      error = RESERVATION_FAILED;
    }
  }
  else if(!strncmp(operation, "cancel", 6)) {
    fprintf(stderr, "About to call _reservation_cancel()\n");
    error = globus_gara_end2end_reservation_cancel(reservation_handle);
    fprintf(stderr, "error = %d\n", error);
  }
  if(!error && !strncmp(operation, "create", 6)) {
    size = strlen(reservation_handle);
    write(1, &size, sizeof(int));
    fprintf(stdout, "%s", reservation_handle);
  }
 
  return error;
}

/***********************************************************
 *
 * Function: make_reservation_rsl
 * Purpose:  Construct an RSL string that we can use to
 *           create or modify a reseration.
 *
 ***********************************************************/
static char *make_reservation_rsl(res_params_t *params) {
  char *rsl_string;

  rsl_string = (char *)malloc(4096);
  if (rsl_string != GLOBUS_NULL)
  {
    sprintf(rsl_string,
	    "&(reservation-type=network) (start-time=%d) (duration=%d) (endpoint-a=%s) (endpoint-b=%s) (bandwidth=%d) (protocol=%s)", 
	    //"&(reservation-type=network) (start-time=%d) (duration=%d) (endpoint-a=%s) (endpoint-b=%s) (bandwidth=%d) (protocol=%s) (auth-data=%s)", 
	    (int) time(NULL) + params->s_time_offset, params->duration, 
	    params->host1, params->host2, params->bandwidth, params->protocol);
	    //params->host1, params->host2, params->bandwidth, params->protocol,
            //"garbage");
//"287f42c19122ec08ddc679ba259070f9/x509:akenti_members/x509:akenti_friends/x509:akenti_src/university.a.g:student.nrgs.a.g/x509:source/x509:friends/mdw:afs33src/itd.login:subscribers_a/university.a.g:student.elgregna.a.g/university.a.g:student.elgregpe.a.g/university.a.g:student.reg.a.g/university.a.g:employee.a.g/university.a.g:temp.a.g/itd.umich:afssource/itd.umich:friends");
//"287f42c19122ec08ddc679ba259070f9/x509:akenti_members/x509:akenti_friends/x509:akenti_src/university.a.g:student.nrgs.a.g/x509:source/x509:friends/mdw:afs33src/itd.login:subscribers_a/university.a.g:student.elgregna.a.g/university.a.g:student.elgregpe.a.g/university.a.g:student.reg.a.g/university.a.g:employee.a.g/university.a.g:temp.a.g");
}
  return rsl_string;
}

/***********************************************************
 *
 * Function: check_status
 * Purpose:  
 *
 ***********************************************************/
static void check_status(char *reservation_handle) {
  int status;
  int error;

  error = globus_gara_end2end_reservation_status(
              reservation_handle, &status);
  if (error)
  {
    fprintf(stderr, 
        "Error %d when checking status for %s\n",
	error, reservation_handle);
  }
  else
  {
    fprintf(stderr,
        "Status is %d for %s\n", status, reservation_handle);
  }
  return;
}

/***********************************************************
 *
 * Function: callback
 * Purpose:  
 *
 ***********************************************************/
static void callback(char *reservation_handle, int event, 
		     void  *user_parameter) 
{
  fprintf(stderr, "\n");
  fprintf(stderr, "Received callback:\n");
  fprintf(stderr, "    Handle: %s\n", reservation_handle);
  fprintf(stderr, "    Event:  %d\n\n", event);

  return;
}

/***********************************************************
 *
 * Function: process_arguments
 * Purpose:  Examine the command line arguments.
 *
 ***********************************************************/
static void process_arguments(res_params_t  *params) {

  //params->host1 = ;
  //params->host2 = ;
  //params->s_time_offset = ;
  //params->duration = ;
  //params->bandwidth = ;
  //params->protocol = ;

  return;
}

