/* apache includes */
#include "ap_config.h"
#include "httpd.h"
#include "http_config.h"
#include "http_conf_globals.h"
#include "http_protocol.h"
#include "http_request.h"
#include "http_main.h"
#include "http_core.h"
#include "http_log.h"

#include <openssl/x509.h>
#include <openssl/ssl.h>

/* #define QOS_HEADER \
"<HTML><HEAD><TITLE>Authenticated QoS Reservation</TITLE></HEAD>\
<BODY BGCOLOR=\"#FFFFFF\" TEXT=\"#000000\">\
<H3>Authenticated QoS Reservation</H3>\
<P>This reservation interface to \
<A HREF=\"http://www-fp.mcs.anl.gov/qos/\">GARA</A>\
 is protected by mutual SSL authentication.<P>\
<P><HR><P><UL><P>\n" */

#define QOS_HEADER "<HTML><BODY BGCOLOR=\"#FFFFFF\" TEXT=\"#000000\">\n"
 
/*#define QOS_FOOTER \
"<P></UL><HR>page maintained by: <A HREF=\"mailto:someone@citi.umich.edu\">someone@citi.umich.edu</A></BODY></HTML>\n"                                          */

#define QOS_FOOTER "</BODY></HTML>\n"    
#define EMAIL_SIZE    100
#define CLIENTID_SIZE 9                                                       
#define NO_ERROR 0

#define AUTHORIZATION_FAILED 2 
#define RESERVATION_FAILED   3

typedef struct {
    char *prog;
    char *x509_cert_dir;
    char *x509_cert_file;
    char *x509_user_cert;
    char *x509_user_key;
    char *x509_user_proxy_prefix;
    char *tkt_cache_prefix;
} GARASrvConfigRec;
 
extern module MODULE_VAR_EXPORT e2egara_module;
                        
#define cfgMerge(el,unset)  new->el = add->el == unset ? base->el : add->el
#define mySrvConfig(srv) (GARASrvConfigRec *)ap_get_module_config(srv->module_config,  &e2egara_module) 
                                                 
static int get_client_identity(pool *p, X509 *, char **);
static int get_arguments(char **x509_user_proxy, char *id, request_rec *r);

void *gara_config_server_create(pool *p, server_rec *s) {
  GARASrvConfigRec *sc;
 
  sc = ap_palloc(p, sizeof(GARASrvConfigRec));
  sc->prog = NULL;
  sc->x509_cert_dir = NULL;
  sc->x509_cert_file = NULL;
  sc->x509_user_cert = NULL;
  sc->x509_user_key = NULL;
  sc->x509_user_proxy_prefix = NULL;
  return sc;
} 

void *gara_config_server_merge(pool *p, void *basev, void *addv) {
   GARASrvConfigRec *base = (GARASrvConfigRec *)basev;
   GARASrvConfigRec *add = (GARASrvConfigRec *)add;
   GARASrvConfigRec *new = (GARASrvConfigRec *)
     ap_palloc(p, sizeof(GARASrvConfigRec));
 
   cfgMerge(prog, NULL);
   cfgMerge(x509_cert_dir, NULL);
   cfgMerge(x509_cert_file, NULL);
   cfgMerge(x509_user_cert, NULL);
   cfgMerge(x509_user_key, NULL);
   cfgMerge(x509_user_proxy_prefix, NULL);
 
   return new;
}

const char *gara_cmd_GARA_end2end(cmd_parms *cmd, void *mconfig, char *arg) {
  GARASrvConfigRec *sc = mySrvConfig(cmd->server);
  struct stat sb;
 
  sc->prog = ap_server_root_relative(cmd->pool, arg);
  if((sc->prog == NULL) || (stat(sc->prog, &sb) != 0) ||
     !S_ISREG(sb.st_mode) || (sb.st_mode <= 0))
    return ap_pstrcat(cmd->pool, "GARA_end2end: file '", sc->prog,
                      "' not exists or empty", NULL);
  return NULL;
}

const char *gara_cmd_X509CertDir(cmd_parms *cmd, void *mconfig, char *arg) {
  GARASrvConfigRec *sc = mySrvConfig(cmd->server);
  struct stat sb;
 
  sc->x509_cert_dir = ap_server_root_relative(cmd->pool, arg);
  if((sc->x509_cert_dir == NULL) || (stat(sc->x509_cert_dir, &sb) != 0) ||
     !S_ISDIR(sb.st_mode) || (sb.st_mode <= 0))
    return ap_pstrcat(cmd->pool, "X509CertDir: file '", sc->x509_cert_dir,
                      "' not exists or empty", NULL);
  return NULL;
 
}

const char *gara_cmd_X509CertFile(cmd_parms *cmd, void *mconfig, char *arg) {
  GARASrvConfigRec *sc = mySrvConfig(cmd->server);
  struct stat sb;
 
  sc->x509_cert_file = ap_server_root_relative(cmd->pool, arg);
  if((sc->x509_cert_file == NULL) || (stat(sc->x509_cert_file, &sb) != 0) ||
     !S_ISREG(sb.st_mode) || (sb.st_mode <= 0))
    return ap_pstrcat(cmd->pool, "X509CertFile: file '", sc->x509_cert_file,
                      "' not exists or empty", NULL);
  return NULL;
}

const char *gara_cmd_X509UserCert(cmd_parms *cmd, void *mconfig, char *arg) {
  GARASrvConfigRec *sc = mySrvConfig(cmd->server);
  int size = strlen(arg);
 
  sc->x509_user_cert = ap_pcalloc(cmd->pool, size+1);
  strncpy(sc->x509_user_cert, arg, size);
 
  return NULL;
}
 
const char *gara_cmd_X509UserKey(cmd_parms *cmd, void *mconfig, char *arg) {
  GARASrvConfigRec *sc = mySrvConfig(cmd->server);
  int size = strlen(arg);
 
  sc->x509_user_key=ap_pcalloc(cmd->pool, size+1);
  strncpy(sc->x509_user_key, arg, size);
 
  return NULL; 
}

const char *gara_cmd_X509UserProxy(cmd_parms *cmd, void *mconfig, char *arg) {
  GARASrvConfigRec *sc = mySrvConfig(cmd->server);
  int size = strlen(arg);
 
  sc->x509_user_proxy_prefix=ap_pcalloc(cmd->pool, size+1);
  strncpy(sc->x509_user_proxy_prefix, arg, size);
 
  return NULL;
}                                                                                                   
static int get_client_identity(pool *p, X509 *cert, char **name) {
 
  X509_NAME *client=NULL;
  X509_NAME_ENTRY *entry=NULL;
  int i, n, entries;
  char *tmp;
 
  tmp = (unsigned char *)ap_pcalloc(p, EMAIL_SIZE);
  (*name) = (unsigned char *)ap_pcalloc(p, CLIENTID_SIZE);
 
  client = X509_get_subject_name(cert);
  X509_NAME_get_text_by_NID(client, NID_pkcs9_emailAddress, tmp,
                            EMAIL_SIZE);
  sscanf(tmp, "%[^@]", (*name));
  (*name)[strlen(*name)] = '\0';
  return OK;
} 
                                                                             
static int get_arguments(char **x509_user_proxy, char *client_id,
                         request_rec *r) {
  GARASrvConfigRec *sc = mySrvConfig(r->server);
                                 
  (*x509_user_proxy) = (char *)
     ap_pcalloc(r->pool, strlen(client_id)+strlen(sc->x509_user_proxy_prefix)+1);
  sprintf((*x509_user_proxy), "%s%s", sc->x509_user_proxy_prefix, client_id);
 
  fprintf(stderr, "%s\n", (*x509_user_proxy));
 
  return 1;
}  

static void find_reservation_handle(char *handle_id, pool *p,
                                    char **reservation_handle) {

  FILE *f=fopen("/tmp/handles", "r");
  char *line=NULL;
  char *tmp=NULL;
  char *b;
 
  line = (char *)ap_pcalloc(p, 1024);
  tmp  = (char *)ap_pcalloc(p, 128);
  while(1) {
    fgets(line, 1024, f);
    sscanf(line, "%s", tmp);
    if(!strncmp(tmp, handle_id, strlen(handle_id))) {
      *reservation_handle = (char *)ap_palloc(p, 1024);
      b = line + strlen(tmp) + 1;
      strncpy(*reservation_handle, b, 1024);     
      break;
    } 
    if(line[0] == '\0' || line[0] == EOF) break;
  } 
  fclose(f);
}

static void save_reservation_handle(long id, char *reservation_handle) {
  FILE *f=fopen("/tmp/handles", "a");
  fprintf(f, "%ld %s\n", id, reservation_handle);
  fclose(f);
}

static void remove_reservation_handle(long id) {
} 
static int e2egara_handler(request_rec *r) {
  GARASrvConfigRec *sc = mySrvConfig(r->server);
  char *client_id;
  SSL *s = ap_ctx_get(r->connection->client->ctx, "ssl");                         int pid, retval, status = 0;
  const char *host1 = NULL, *host2 = NULL, *duration = NULL, 
             *bandwidth = NULL, *protocol = NULL, *offset = NULL, 
             *type = NULL, *handle_id = NULL;
  char *x509_user_proxy = NULL; 
  table *t = r->subprocess_env;
  int p1[2];
  int close_on_exec_read_fd, close_on_exec_write_fd;                            
  int handle_size;
  char *reservation_handle = NULL;
  long id = 0;

  if(s->session->peer == NULL) {
    fprintf(stderr, "unable to get client's certificate\n");
    return HTTP_INTERNAL_SERVER_ERROR; 
  }

  if(get_client_identity(r->pool, s->session->peer, &client_id) < 0) {
    fprintf(stderr, "wasn't able to get client's identity from the cert\n");
    return HTTP_INTERNAL_SERVER_ERROR;
  }                                                                             

  type = (const char *)ap_table_get(t, "E2E_TYPE");                             
  if(!strncmp(type, "create", 6)) {
    host1 = (const char *)ap_table_get(t, "E2E_HOST1");
    host2 = (const char *)ap_table_get(t, "E2E_HOST2");
    offset = (const char *)ap_table_get(t, "E2E_OFFSET");
    bandwidth = (const char *)ap_table_get(t, "E2E_BANDWIDTH");
    duration = (const char *)ap_table_get(t, "E2E_DURATION");
    protocol = (const char *)ap_table_get(t, "E2E_PROTOCOL");
    fprintf(stderr, "mod_e2egara: type = %s\nhost1 = %s\nhost2 = %s\nbandwidth = %s\noffset = %s\nduration = %s\nprotocol = %s\n", type, host1, host2, bandwidth, offset, duration, protocol);
  }
  else {
    handle_id = (const char *)ap_table_get(t, "E2E_HANDLE_ID");
  }

  if(!get_arguments(&x509_user_proxy, client_id, r)) goto err;
  
  if(pipe(p1) != 0) {
    perror("pipe failed\n");
    return -1;
  }
  close_on_exec_read_fd = p1[0];
  close_on_exec_write_fd = p1[1];                                               

  if((pid = fork()) < 0) {
    fprintf(stderr, "fork failed\n");
    goto err;
  } else if(pid == 0) { /*child */
    close(p1[0]);
    if(dup2(p1[1], STDOUT_FILENO) != STDOUT_FILENO)
      perror("dup2 failed\n");                                                  

    fprintf(stderr, "%d: child process calling e2egara\n", getpid());
    if(!strncmp(type, "cancel", 6)) {
      if(handle_id == NULL) {
        fprintf(stderr, "Reservation number is missing\n");
        return HTTP_INTERNAL_SERVER_ERROR;
      }
      find_reservation_handle(handle_id, r->pool, &reservation_handle);
      if(reservation_handle == NULL) {
        fprintf(stderr, "Unable to find the reservation handle for %s\n",
                handle_id);
        return HTTP_INTERNAL_SERVER_ERROR;
      }
      reservation_handle[strlen(reservation_handle)-1] = '\0';
      if(execl(sc->prog, sc->prog, sc->x509_cert_dir, sc->x509_cert_file,
             sc->x509_user_cert, sc->x509_user_key, x509_user_proxy,
             type, reservation_handle, NULL) < 0) {
        perror("exec failed\n");
        goto err;
      }       
    } else { 
      if(execl(sc->prog, sc->prog, sc->x509_cert_dir, sc->x509_cert_file, 
             sc->x509_user_cert, sc->x509_user_key, x509_user_proxy, 
             type, host1, host2, offset, bandwidth, duration, protocol, 
             NULL) < 0) {
        perror("exec failed\n");
        goto err;
      }
    }
  } else { /* parent */
    close(p1[1]);                                           

    fprintf(stderr, "%d: parent process waiting for child %d\n", 
            getpid(), pid);
    waitpid(pid, &status, 0);
    fprintf(stderr, "%d: child is done, exit status = %d %d\n", 
            getpid(), status, WEXITSTATUS(status));

    if(!status && !strncmp(type, "create", 6)) {
      read(p1[0], &handle_size, sizeof(int));
      if(handle_size < 0) return -1;
 
      reservation_handle = (char *)ap_palloc(r->pool, handle_size);
      read(p1[0], reservation_handle, handle_size);
      reservation_handle[handle_size] = '\0'; 
     
      fprintf(stderr, "reservation_handle = %s\n", reservation_handle);
      /* add the reservation handle to a table indexed by an unique number */
      id = random();
      save_reservation_handle(id, reservation_handle);
      fprintf(stderr, "unique id: %ld\n", id);
    }
  }

  r->content_type = "text/html";
  //ap_send_http_header(r);
  //ap_rputs(QOS_HEADER, r);
  //ap_rputs(QOS_FOOTER, r);
  //return OK;

  switch(WEXITSTATUS(status)) {
  case NO_ERROR:
    if(!strncmp(type, "create", 6))  
      ap_rprintf(r, "<P><FONT COLOR=red>Reservation has suceeded. Reservation number: %ld</FONT></P>", id);
    else 
      ap_rprintf(r, "<P><FONT COLOR=red>Reservation number: %s has been cancelled</FONT></P>", handle_id);                                                    
    break;
  case AUTHORIZATION_FAILED:
    if(!strncmp(type, "create", 6)) 
      ap_rputs("<P><FONT COLOR=red>Error: unable to make a reservation: authorization failed.</FONT></P>", r);
    else
      ap_rprintf(r, "<P><FONT COLOR=red>Error: unable to cancel a reservation number %s: authorization failed.</FONT></P>", handle_id);                                                  
    break;                                                                      
  default:
    ap_rputs("<P><FONT COLOR=red>Error: unable to make a reservation.</FONT></P>", r);
    break;
  }
  
  ap_rputs(QOS_FOOTER, r);
  //return DECLINED;
  return OK;

err:
  ap_rputs(QOS_FOOTER, r);
  return HTTP_INTERNAL_SERVER_ERROR;
}

static command_rec gara_config_cmds[] = {
  {"GARA_end2end", gara_cmd_GARA_end2end, NULL, RSRC_CONF, TAKE1,
   "specify the filename of the e2e executable"},
  {"GX509_CERT_DIR", gara_cmd_X509CertDir, NULL, RSRC_CONF, TAKE1,
   "specify the path for X509_CERT_DIR"},
  {"GX509_CERT_FILE", gara_cmd_X509CertFile, NULL, RSRC_CONF, TAKE1,
   "specify the filename for X509_CERT_FILE"},
  {"GX509_USER_CERT", gara_cmd_X509UserCert, NULL, RSRC_CONF, TAKE1,
   "specify the filename for X509_USER_CERT"},
  {"GX509_USER_KEY", gara_cmd_X509UserKey, NULL, RSRC_CONF, TAKE1,
   "specify the filename for X509_USER_KEY"},
  {"GX509_USER_PROXY", gara_cmd_X509UserProxy, NULL, RSRC_CONF, TAKE1,
   "specify the prefix for X509_USER_PROXY"},
  {NULL},
}; 
static handler_rec e2egara_handlers[] = {
  {"*/*", e2egara_handler},
  {NULL}
};
 
module MODULE_VAR_EXPORT e2egara_module = {                  
  STANDARD_MODULE_STUFF,
  NULL, /* initializer */
  NULL, /* create per-dir config */
  NULL, /* merge per-dir config */
  gara_config_server_create, /* create per-server config */
  gara_config_server_merge, /* merge per-server config */
  gara_config_cmds, /* command table */
  e2egara_handlers, /* [#8] handlers */
  NULL, /* [#1] file name translation */
  NULL, /* [#4] check user id */
  NULL, /* [#5] check auth */
  NULL, /* [#3] check access */
  NULL, /* [#6] type checker */ 
  NULL, /* [#7] prerun fixups */ 
  NULL, /* [#9] logger */
  NULL, /* [#2] header parser */
  NULL, /* child init */
  NULL, /* child exit */
  NULL,  /* [#0] post read request */
};
